function K = imm_kpca_kernel_matrix(X_train, X_test, p)

% Computes the kernel matrix K
%
% INPUT:
%         p.X_train: training examples
%         p.X_test:  test examples
%         p.kernel:  'rbf', 'poly', 'sigmoid', 'linear'
%         p.c:       vector with kernel parameters:
%                    linear:  k(x,y)=x'*y+c(1)
%                    rbf:     k(x,y)=exp(-0.5*dist(x,y)^2/c(1)^2)
%                    poly:    k(x,y)=(x'*y).^c(1)
%                    sigmoid: k(x,y)=tanh(c(1)*x'*y+c(2))
% OUTPUT:
%         K:         kernel matrix
%
% Kasper W Jorgensen, IMM-DTU, 2012

if ~isfield(p,'kernel')
    p.kernel='rbf';
end

switch lower(p.kernel)
    case 'linear'        
        K = X_train'*X_test;
    case 'poly'
        d = p.c(1);
        innerprod=X_train'*X_test;
        K = innerprod.^d;
    case 'sigmoid'
        a = p.c(1);
        c = p.c(2);
        innerprod=X_train'*X_test;
        K = tanh(a*innerprod+c);
    otherwise % defaults to 'rbf'
        % get the squared euclidean distance
        d2 = imm_kpca_euclidean_dist(X_train',X_test');
        c  = p.c(1);
        K  = exp(-0.5*d2/c^2);
end
