function p = imm_kpca_preimg_kwok(p)

% Implementation of pre-image algorithm presented in:
%    J. T. Kwok and I. W. Tsang, The pre-image problem in kernel
%    methods, Neural Networks, IEEE Transactions on, vol. 15, no. 6,
%    pp. 15171525, November 2004

% number of nearest neighbors to use
if ~isfield(p,'nn')
    nn = 10;
else
    nn = p.nn;
end

K_test  = p.K_test;
X_test  = p.X_test;
X_train = p.X_train;

One = p.One;
K = p.K;
H = p.H;
M = p.M;
N = p.N;
c = p.c;

N_test = size(X_test,2);
const = (One'*K*One) + 1;

for n=1:N_test
    
    k = K_test(:,n);
    
    % center projection
    kx_c  = H*(k - K*One);
    gamma = kx_c' * M;
    
    kKOne =(k - K*One);
    
    d_hat2=zeros(1,N);
    for i=1:N
        kxi = K(:,i);
        tmp_part  = (kKOne - 2*kxi)'*H';
        d_hat2(i) = tmp_part*gamma' - 2*One'*kxi;
    end
    d_hat2=d_hat2+const;
    
    % calc input space distances
    d2 = -2 * c^2 * log(1 - 0.5*d_hat2);
    
    % sort to get the nearest nn samples
    [sorted_d2 indx] = sort(d_hat2);
    X_tmp  = X_train(:,indx(1:nn));
    d2     = d2(indx(1:nn))';    
    
    H_tmp   = eye(nn) - ones(nn)/nn;
    [U,S,V] = svd(X_tmp*H_tmp);
    
    Z = S*V';
    
    % zero out very small numbers
    Z=(abs(Z)>10e-10).*Z;
    
    d02 = sum(Z.^2)';
    
    z = -0.5*pinv(Z')*(d2-d02);
    x = U*z + sum(X_tmp,2)/nn;
    p.X_preimg(:,n) = x;
    
end

