function p = imm_kpca_preimg_mika(p)

% USAGE:
%   p = imm_kpca_preimg_mika(p)
%
% Implementation of the pre-image algorithm presented in:
%    S. Mika, B. Scholkopf, A. Smola, K.-R. Muller, M. Scholz, and G. Ratsch,
%    Kernel pca and de-noising in feature spaces, in Advances in Neural
%    Information Processing Systems 11. MIT Press, 1999, pp. 536542. 

max_iter = 1000;
conv_criteria = 10e-10;

X_test  = p.X_test;
K_test  = p.K_test;
X_train = p.X_train;
K = p.K;
H = p.H;
M = p.M;
N = p.N;
One = p.One;

N_test = size(X_test,2);

for n=1:N_test
    
    % test points
    k = K_test(:,n);
    
    if p.random_start==1
        % create random starting point
        x = X_test(:,n)+randn(size(X_test,1),1);        
    else 
        % else start in noise data points
        x = X_test(:,n);
    end
    
    % center kernel matrix
    kx_c  = H*(k - K*One);
    
    gamma = kx_c' * M;
    gamma_hat = gamma + (1 - sum(gamma))/N;
    
    for iter=1:max_iter                
        
        k          = imm_kpca_kernel_matrix(X_train,x,p);
        kgh        = k.*gamma_hat';
        sum_kgh    = sum(kgh);
        x_new      = X_train*kgh/sum_kgh;
        dist(iter) = sum((x_new-x).^2);
        x          = x_new;
        if dist(iter)<conv_criteria,           
            break;
        end
    end
    
    p.X_preimg(:,n)    = x;        
end
