function p = imm_kpca_preimg_rathi(p)

% Implementation of pre-image algorithm presented in:
%    Yogesh Rathi, Samuel Dambreville,  Allen Tannenbaum
%    'Statistical Shape Analysis using Kernel PCA', in SPIE, Electronic
%    Imaging, 2006

X_test  = p.X_test;
K_test  = p.K_test;
X_train = p.X_train;

K = p.K;
H = p.H;
M = p.M;
N = p.N;
One = p.One;

N_test = size(X_test,2);

%----------------------------------
% pre image
p.X_preimg = zeros(size(X_test));

% for each test point
for n=1:N_test
    
    % test point in feature space
    kx = K_test(:,n);
    
    % calc distance from test point to all trainings projections
    kxKone    = kx + K*One;
    kx_c      = H*(kx - K*One);
    hmhkxkone = H'*M*kx_c;
    onekone   = (One'*K*One);
    dist      = zeros(1,N);
    for i=1:N
        kxi = p.K(:,i);
        dist(i) = (kxKone - 2*kxi)'*hmhkxkone + onekone + 1 - 2*One'*kxi;
    end
    
    %---------------------------------------------------------
    % calc gamma
    gamma = kx_c' * M;
    gamma_hat = gamma + (1 - sum(gamma))/N;
    
    k_hat   = (2 - dist)/2;
    kgh     = k_hat.*gamma_hat;
    sum_kgh = sum(kgh);
    
    x_new   = X_train*kgh'/sum_kgh;
    p.X_preimg(:,n) = x_new;        
    
end
